"""
Money class unittests
"""
import unittest

from money import Money
from . import mixins


class TestMoneyInstantiation(mixins.InstantiationMixin, unittest.TestCase):
    def setUp(self):
        self.MoneyClass = Money


class TestMoneyClass(mixins.ClassMixin, unittest.TestCase):
    def setUp(self):
        self.money = Money('2.99', 'XXX')


class TestMoneyRepresentations(mixins.RepresentationsMixin, unittest.TestCase):
    def setUp(self):
        self.money = Money('1234.567', 'XXX')


class TestMoneyFormatting(mixins.FormattingMixin, unittest.TestCase):
    def setUp(self):
        self.money = Money('-1234.567', 'USD')


class TestMoneyParser(mixins.ParserMixin, unittest.TestCase):
    def setUp(self):
        self.MoneyClass = Money


class TestMoneyNumericOperations(
        mixins.NumericOperationsMixin,
        unittest.TestCase):
    def setUp(self):
        self.MoneyClass = Money


class TestMoneyUnaryOperationsReturnNew(
        mixins.UnaryOperationsReturnNewMixin,
        unittest.TestCase):
    def setUp(self):
        self.money = Money('2.99', 'XXX')


class TestMoneyLeftmostTypePrevails(
        mixins.LeftmostTypePrevailsMixin,
        unittest.TestCase):
    def setUp(self):
        self.MoneyClass = Money
        self.money = self.MoneyClass('2.99', 'XXX')
        self.MoneySubclass = type('MoneySubclass', (self.MoneyClass,), {})
        self.other_money = self.MoneySubclass('2.99', 'XXX')



"""
Money exchange unittests
"""

from decimal import Decimal
import unittest


from money import Money, XMoney, xrates
from money.exchange import SimpleBackend
from money.exceptions import ExchangeBackendNotInstalled
from money.exceptions import ExchangeRateNotFound


class TestExchangeRatesSetup(unittest.TestCase):
    def setUp(self):
        xrates.uninstall()

    def test_register(self):
        self.assertFalse(xrates)
        self.assertIsNone(xrates.backend_name)
        xrates.install('money.exchange.SimpleBackend')
        self.assertEqual(xrates.backend_name, 'SimpleBackend')

    def test_register_class(self):
        self.assertFalse(xrates)
        self.assertIsNone(xrates.backend_name)
        xrates.install(SimpleBackend)
        self.assertEqual(xrates.backend_name, 'SimpleBackend')

    def test_register_instance(self):
        self.assertFalse(xrates)
        self.assertIsNone(xrates.backend_name)
        xrates.install(SimpleBackend())
        self.assertEqual(xrates.backend_name, 'SimpleBackend')

    def test_unregister(self):
        xrates.install('money.exchange.SimpleBackend')
        self.assertEqual(xrates.backend_name, 'SimpleBackend')
        xrates.uninstall()
        self.assertFalse(xrates)
        self.assertIsNone(xrates.backend_name)

    def test_no_backend_name(self):
        self.assertIsNone(xrates.backend_name)

    def test_no_backend_base(self):
        with self.assertRaises(ExchangeBackendNotInstalled):
            xrates.base

    def test_no_backend_get_rate(self):
        with self.assertRaises(ExchangeBackendNotInstalled):
            xrates.rate('XXX')

    def test_no_backend_get_quotation(self):
        with self.assertRaises(ExchangeBackendNotInstalled):
            xrates.quotation('XXX', 'YYY')

    def test_multiple_xrates(self):
        xrates.install('money.exchange.SimpleBackend')
        self.assertTrue(xrates)
        xrates.base = 'XXX'
        xrates.setrate('AAA', Decimal('2'))

        from money.exchange import ExchangeRates
        another = ExchangeRates()
        another.install('money.exchange.SimpleBackend')
        self.assertTrue(another)
        another.base = 'XXX'
        another.setrate('AAA', Decimal('100'))

        self.assertEqual(xrates.rate('AAA'), Decimal('2'))
        self.assertEqual(another.rate('AAA'), Decimal('100'))


class TestSimpleBackend(unittest.TestCase):
    def setUp(self):
        xrates.install('money.exchange.SimpleBackend')
        xrates.base = 'XXX'
        xrates.setrate('AAA', Decimal('2'))
        xrates.setrate('BBB', Decimal('8'))

    def test_base_property(self):
        self.assertEqual(xrates.base, 'XXX')

    def test_rate(self):
        self.assertEqual(xrates.rate('XXX'), Decimal('1'))
        self.assertEqual(xrates.rate('AAA'), Decimal('2'))
        self.assertEqual(xrates.rate('BBB'), Decimal('8'))

    def test_quotation(self):
        self.assertEqual(xrates.quotation('XXX', 'XXX'), Decimal('1'))
        self.assertEqual(xrates.quotation('XXX', 'AAA'), Decimal('2'))
        self.assertEqual(xrates.quotation('XXX', 'BBB'), Decimal('8'))
        self.assertEqual(xrates.quotation('AAA', 'XXX'), Decimal('0.5'))
        self.assertEqual(xrates.quotation('AAA', 'AAA'), Decimal('1'))
        self.assertEqual(xrates.quotation('AAA', 'BBB'), Decimal('4'))
        self.assertEqual(xrates.quotation('BBB', 'XXX'), Decimal('0.125'))
        self.assertEqual(xrates.quotation('BBB', 'AAA'), Decimal('0.25'))
        self.assertEqual(xrates.quotation('BBB', 'BBB'), Decimal('1'))

    def test_unavailable_rate_returns_none(self):
        self.assertIsNone(xrates.rate('ZZZ'))

    def test_unavailable_quotation_returns_none(self):
        self.assertIsNone(xrates.quotation('YYY', 'ZZZ'))

    def test_money_conversion(self):
        self.assertEqual(Money('10', 'AAA').to('BBB'), Money('40', 'BBB'))
        self.assertEqual(Money('10', 'BBB').to('AAA'), Money('2.5', 'AAA'))

    def test_base_not_set_warning(self):
        xrates.uninstall()
        xrates.install('money.exchange.SimpleBackend')
        with self.assertRaises(Warning):
            xrates.setrate('AAA', Decimal('2'))


class ConversionMixin(object):
    def test_unavailable_backend_conversion_error(self):
        xrates.uninstall()
        with self.assertRaises(ExchangeBackendNotInstalled):
            self.MoneyClass('2', 'AAA').to('BBB')

    def test_unavailable_rate_conversion_error(self):
        xrates.install('money.exchange.SimpleBackend')
        with self.assertRaises(ExchangeRateNotFound):
            self.MoneyClass('2', 'AAA').to('BBB')


class TestMoneyConversion(ConversionMixin, unittest.TestCase):
    MoneyClass = Money


class TestXMoneyConversion(ConversionMixin, unittest.TestCase):
    MoneyClass = XMoney



"""
XMoney class unittests
"""
from decimal import Decimal
import unittest

from money import XMoney, xrates
from . import mixins


class TestXMoneyInstantiation(mixins.InstantiationMixin, unittest.TestCase):
    def setUp(self):
        self.MoneyClass = XMoney


class TestXMoneyClass(mixins.ClassMixin, unittest.TestCase):
    def setUp(self):
        self.money = XMoney('2.99', 'XXX')


class TestXMoneyRepresentations(
        mixins.RepresentationsMixin,
        unittest.TestCase):
    def setUp(self):
        self.money = XMoney('1234.567', 'XXX')


class TestXMoneyFormatting(mixins.FormattingMixin, unittest.TestCase):
    def setUp(self):
        self.money = XMoney('-1234.567', 'USD')


class TestXMoneyParser(mixins.ParserMixin, unittest.TestCase):
    def setUp(self):
        self.MoneyClass = XMoney


class TestXMoneyNumericOperations(
        mixins.NumericOperationsMixin,
        unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        xrates.install('money.exchange.SimpleBackend')
        xrates.base = 'XXX'
        xrates.setrate('AAA', Decimal('2'))
        xrates.setrate('BBB', Decimal('8'))

    @classmethod
    def tearDownClass(cls):
        xrates.uninstall()

    def setUp(self):
        self.MoneyClass = XMoney
        self.x = XMoney('10', 'XXX')
        self.a = XMoney('10', 'AAA')
        self.b = XMoney('10', 'BBB')
        self.ax = XMoney('20', 'AAA')
        self.bx = XMoney('80', 'BBB')

    def test_add_money_different_currency(self):
        self.assertEqual(self.a + self.b, XMoney('12.5', 'AAA'))
        self.assertEqual(self.b + self.a, XMoney('50', 'BBB'))

    def test_sub_money_different_currency(self):
        self.assertEqual(self.a - self.b, XMoney('7.5', 'AAA'))
        self.assertEqual(self.b - self.a, XMoney('-30', 'BBB'))

    def test_truediv_money_different_currency(self):
        self.assertEqual(self.a / self.b, Decimal('4'))
        self.assertEqual(self.b / self.a, Decimal('0.25'))

    def test_floordiv_money_different_currency(self):
        self.assertEqual(self.a // self.b, Decimal('4'))
        self.assertEqual(self.b // self.a, Decimal('0'))

    def test_divmod_money_different_currency(self):
        whole, remainder = divmod(self.a, self.b)
        self.assertEqual(whole, Decimal('4'))
        self.assertEqual(remainder, Decimal('0'))
        whole, remainder = divmod(self.b, self.a)
        self.assertEqual(whole, Decimal('0'))
        self.assertEqual(remainder, Decimal('10'))


class TestXMoneyUnaryOperationsReturnNew(
        mixins.UnaryOperationsReturnNewMixin,
        unittest.TestCase):
    def setUp(self):
        self.money = XMoney('2.99', 'XXX')


class TestXMoneyLeftmostTypePrevails(
        mixins.LeftmostTypePrevailsMixin,
        unittest.TestCase):
    def setUp(self):
        self.MoneyClass = XMoney
        self.money = self.MoneyClass('2.99', 'XXX')
        self.MoneySubclass = type('MoneySubclass', (self.MoneyClass,), {})
        self.other_money = self.MoneySubclass('2.99', 'XXX')



"""
Money exchange related classes and ``xrates`` API entry point.
"""

import abc
import decimal
import importlib

from .exceptions import ExchangeBackendNotInstalled


class BackendBase(object, metaclass=abc.ABCMeta):
    """Abstract base class API for exchange backends"""

    @property
    @abc.abstractmethod
    def base(self):
        """Return the base currency"""
        return

    @abc.abstractmethod
    def rate(self, currency):
        """Return quotation between the base and another currency"""
        return None

    @abc.abstractmethod
    def quotation(self, origin, target):
        """Return quotation between two currencies (origin, target)"""
        a = self.rate(origin)
        b = self.rate(target)
        if a and b:
            return b / a
        return None


class SimpleBackend(BackendBase):
    def __init__(self):
        self._base = None
        self._rates = {}

    @property
    def base(self):
        return self._base

    @base.setter
    def base(self, currency):
        self._base = currency

    def setrate(self, currency, rate):
        if not self.base:
            raise Warning("set the base first: xrates.base = currency")
        self._rates[currency] = rate

    def rate(self, currency):
        if currency == self.base:
            return decimal.Decimal(1)
        return self._rates.get(currency, None)

    def quotation(self, origin, target):
        return super().quotation(origin, target)


class ExchangeRates(object):
    def __init__(self):
        self._backend = None

    def __bool__(self):
        return bool(self._backend)

    def install(self, backend="money.exchange.SimpleBackend"):
        """Install an exchange rates backend using a python path string"""
        if isinstance(backend, str):
            path, name = backend.rsplit(".", 1)
            module = importlib.import_module(path)
            backend = getattr(module, name)()
        elif isinstance(backend, type):
            backend = backend()
        if not isinstance(backend, BackendBase):
            raise TypeError(
                "backend '{}' is not a subclass of "
                "money.xrates.BackendBase".format(backend)
            )
        self._backend = backend

    def uninstall(self):
        """Uninstall any exchange rates backend"""
        self._backend = None

    @property
    def backend_name(self):
        """Return the class name of the currently installed backend or None"""
        if not self._backend:
            return None
        return self._backend.__class__.__name__

    @property
    def base(self):
        """Return the base currency"""
        if not self._backend:
            raise ExchangeBackendNotInstalled()
        return self._backend.base

    def rate(self, currency):
        """Return quotation between the base and another currency"""
        if not self._backend:
            raise ExchangeBackendNotInstalled()
        return self._backend.rate(currency)

    def quotation(self, origin, target):
        """Return quotation between two currencies (origin, target)"""
        if not self._backend:
            raise ExchangeBackendNotInstalled()
        return self._backend.quotation(origin, target)

    def __getattr__(self, name):
        if self._backend is None:
            raise ExchangeBackendNotInstalled()
        return getattr(self._backend, name)

    def __setattr__(self, name, value):
        if name == "_backend":
            self.__dict__[name] = value
        elif self._backend is None:
            raise ExchangeBackendNotInstalled()
        else:
            setattr(self._backend, name, value)


xrates = ExchangeRates()



"""
Money unittests as mixins for Money and subclasses
"""

from decimal import Decimal
import collections.abc
import pickle
import babel


from money import Money
from money.exceptions import InvalidOperandType
from money.exceptions import CurrencyMismatch


class InstantiationMixin(object):
    def test_new_instance_int_amount(self):
        self.assertEqual(self.MoneyClass(0, 'XXX').amount, Decimal('0.00'))
        self.assertEqual(self.MoneyClass(2, 'XXX').amount, Decimal('2.00'))

    def test_new_instance_decimal_amount(self):
        self.assertEqual(
            self.MoneyClass(
                Decimal('0.00'),
                'XXX').amount,
            Decimal('0.00'))
        self.assertEqual(
            self.MoneyClass(
                Decimal('2.99'),
                'XXX').amount,
            Decimal('2.99'))

    def test_new_instance_float_amount(self):
        self.assertEqual(self.MoneyClass(0.0, 'XXX').amount, Decimal('0.00'))
        self.assertAlmostEqual(
            self.MoneyClass(
                2.99,
                'XXX').amount,
            Decimal('2.99'))

    def test_new_instance_str_amount(self):
        self.assertEqual(self.MoneyClass('0', 'XXX').amount, Decimal('0.00'))
        self.assertEqual(
            self.MoneyClass(
                '2.99',
                'XXX').amount,
            Decimal('2.99'))

    def test_invalid_currency_missing(self):
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99')

    def test_invalid_currency_none(self):
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', None)

    def test_invalid_currency_false(self):
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', False)

    def test_invalid_currency_empty(self):
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', '')

    def test_invalid_currency_code(self):
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', 'XX')
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', '123')
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', 'xxx')
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', '$')
        with self.assertRaises(ValueError):
            self.MoneyClass('2.99', 'US$')

    def test_invalid_amount(self):
        with self.assertRaises(ValueError):
            self.MoneyClass('twenty', 'XXX')


class ClassMixin(object):
    def test_is_money(self):
        self.assertIsInstance(self.money, Money)

    def test_immutable_by_convention(self):
        with self.assertRaises(AttributeError):
            self.money.amount += 1
        with self.assertRaises(AttributeError):
            self.money.currency = 'YYY'

    def test_hashable(self):
        self.assertIsInstance(self.money, collections.abc.Hashable)

    def test_hash_eq(self):
        money_set = set([self.money, self.money])
        self.assertEqual(len(money_set), 1)

    def test_hash_int(self):
        self.assertEqual(type(hash(self.money)), int)

    def test_pickable(self):
        self.assertEqual(pickle.loads(pickle.dumps(self.money)), self.money)

    def test_sqlalchemy_composite_values(self):
        self.assertEqual(
            (self.money.amount,
             self.money.currency),
            self.money.__composite_values__())


class RepresentationsMixin(object):
    def test_repr(self):
        self.assertEqual(repr(self.money), 'XXX 1234.567')

    def test_str(self):
        self.assertEqual(str(self.money), 'XXX 1,234.57')


# RADAR: Python2 (unicode strings u'')
class FormattingMixin(object):
    def test_custom_format_padding(self):
        self.assertEqual(
            self.money.format(
                'en_US',
                u'¤000000.00'),
            u'-$001234.57')

    def test_custom_format_custom_negative(self):
        self.assertEqual(
            self.money.format(
                'en_US',
                u'¤#,##0.00;<¤#,##0.00>'),
            u'<$1,234.57>')

    def test_custom_format_grouping(self):
        self.assertEqual(
            self.money.format(
                'en_US',
                u'¤#,##0.00'),
            u'-$1,234.57')
        self.assertEqual(
            self.money.format(
                'de_DE',
                u'#,##0.00 ¤'),
            u'-1.234,57 $')
        self.assertEqual(self.money.format('en_US', u'¤0.00'), u'-$1234.57')
        self.assertEqual(self.money.format('de_DE', u'0.00 ¤'), u'-1234,57 $')

    def test_custom_format_decimals(self):
        self.assertEqual(
            self.money.format(
                'en_US',
                u'¤0.000',
                currency_digits=False),
            u'-$1234.567')
        self.assertEqual(
            self.money.format(
                'en_US',
                u'¤0',
                currency_digits=False),
            u'-$1235')

    def test_auto_format_locales(self):
        # Test various locales - adapt to current Babel formatting
        us_format = self.money.format('en_US')
        de_format = self.money.format('de_DE')
        es_format = self.money.format('es_CO')

        # Verify the format contains expected elements
        self.assertIn('-$1,234.57', us_format)  # US format
        self.assertIn('1.234,57', de_format)    # German number format
        # Currency code for Colombian locale
        self.assertIn('US$', es_format)
        self.assertIn('1.234,57', es_format)    # Number format

    def test_auto_format_locales_alias(self):
        self.assertEqual(self.money.format('en'), self.money.format('en_US'))
        self.assertEqual(self.money.format('de'), self.money.format('de_DE'))

    def test_auto_format_locale_numeric(self):
        locale = babel.default_locale('LC_NUMERIC')
        if locale is None:
            self.skipTest("No LC_NUMERIC locale environment variable set")
        babel_formatted = babel.numbers.format_currency(
            self.money.amount, self.money.currency, locale=locale)
        self.assertEqual(self.money.format(), babel_formatted)

    def test_auto_format(self):
        # Use explicit locale for consistent testing
        babel_formatted = babel.numbers.format_currency(
            self.money.amount, self.money.currency, locale='en_US')
        self.assertEqual(self.money.format(locale='en_US'), babel_formatted)


class ParserMixin(object):
    def test_loads_repr(self):
        self.assertEqual(
            self.MoneyClass.loads('XXX 2.99'),
            self.MoneyClass(
                '2.99',
                'XXX'))

    def test_loads_missing_currency(self):
        with self.assertRaises(ValueError):
            self.MoneyClass.loads('2.99')

    def test_loads_reversed_order(self):
        with self.assertRaises(ValueError):
            self.MoneyClass.loads('2.99 XXX')

    def test_loads_empty(self):
        with self.assertRaises(ValueError):
            self.MoneyClass.loads('')


class NumericOperationsMixin(object):
    def test_lt(self):
        self.assertTrue(
            self.MoneyClass(
                '2.219',
                'XXX') < self.MoneyClass(
                '2.99',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '-2.99',
                'XXX') < self.MoneyClass(
                '2.99',
                'XXX'))
        self.assertFalse(
            self.MoneyClass(
                '0',
                'XXX') < self.MoneyClass(
                '0',
                'XXX'))

    def test_lt_works_only_with_money(self):
        with self.assertRaises(InvalidOperandType):
            self.MoneyClass(0, 'XXX') < Decimal('0')

    def test_lt_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') < self.MoneyClass(2, 'BBB')

    def test_le(self):
        self.assertTrue(
            self.MoneyClass(
                '2.219',
                'XXX') <= self.MoneyClass(
                '2.99',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '-2.99',
                'XXX') <= self.MoneyClass(
                '2.99',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '0',
                'XXX') <= self.MoneyClass(
                '0',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '2.990',
                'XXX') <= self.MoneyClass(
                '2.99',
                'XXX'))

    def test_le_works_only_with_money(self):
        with self.assertRaises(InvalidOperandType):
            self.MoneyClass(0, 'XXX') <= Decimal('0')

    def test_le_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') <= self.MoneyClass(2, 'BBB')

    def test_eq(self):
        self.assertEqual(
            self.MoneyClass(
                '2', 'XXX'), self.MoneyClass(
                '2', 'XXX'))
        self.assertEqual(
            hash(
                self.MoneyClass(
                    '2', 'XXX')), hash(
                self.MoneyClass(
                    '2', 'XXX')))

        self.assertEqual(
            self.MoneyClass(
                '2.99000', 'XXX'), self.MoneyClass(
                '2.99', 'XXX'))
        self.assertEqual(
            hash(
                self.MoneyClass(
                    '2.99000', 'XXX')), hash(
                self.MoneyClass(
                    '2.99', 'XXX')))

    def test_ne(self):
        self.assertNotEqual(
            self.MoneyClass(
                '0', 'XXX'), self.MoneyClass(
                '2', 'XXX'))
        self.assertNotEqual(
            hash(
                self.MoneyClass(
                    '0', 'XXX')), hash(
                self.MoneyClass(
                    '2', 'XXX')))

        self.assertNotEqual(
            self.MoneyClass(
                '2.99001', 'XXX'), self.MoneyClass(
                '2.99', 'XXX'))
        self.assertNotEqual(
            hash(
                self.MoneyClass(
                    '2.99001', 'XXX')), hash(
                self.MoneyClass(
                    '2.99', 'XXX')))

        self.assertNotEqual(
            self.MoneyClass(
                '2', 'XXX'), self.MoneyClass(
                '2', 'YYY'))
        self.assertNotEqual(
            hash(
                self.MoneyClass(
                    '2', 'XXX')), hash(
                self.MoneyClass(
                    '2', 'YYY')))

    def test_ne_if_not_money(self):
        self.assertNotEqual(self.MoneyClass(0, 'XXX'), Decimal('0'))

    def test_gt(self):
        self.assertTrue(
            self.MoneyClass(
                '2.99',
                'XXX') > self.MoneyClass(
                '2.219',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '2.99',
                'XXX') > self.MoneyClass(
                '-2.99',
                'XXX'))
        self.assertFalse(
            self.MoneyClass(
                '0',
                'XXX') > self.MoneyClass(
                '0',
                'XXX'))

    def test_gt_works_only_with_money(self):
        with self.assertRaises(InvalidOperandType):
            self.MoneyClass(0, 'XXX') > Decimal('0')

    def test_gt_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') > self.MoneyClass(2, 'BBB')

    def test_ge(self):
        self.assertTrue(
            self.MoneyClass(
                '2.99',
                'XXX') >= self.MoneyClass(
                '2.219',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '2.99',
                'XXX') >= self.MoneyClass(
                '-2.99',
                'XXX'))
        self.assertTrue(
            self.MoneyClass(
                '2.99',
                'XXX') >= self.MoneyClass(
                '2.99',
                'XXX'))

    def test_ge_works_only_with_money(self):
        with self.assertRaises(InvalidOperandType):
            self.MoneyClass(0, 'XXX') >= Decimal('0')

    def test_ge_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') >= self.MoneyClass(2, 'BBB')

    def test_bool_true(self):
        self.assertTrue(self.MoneyClass('2.99', 'XXX'))
        self.assertTrue(self.MoneyClass('-1', 'XXX'))

    def test_bool_false(self):
        self.assertFalse(self.MoneyClass('0', 'XXX'))

    def test_add_int(self):
        result = self.MoneyClass('2', 'XXX') + 2
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_add_decimal(self):
        result = self.MoneyClass('2', 'XXX') + Decimal('2')
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_add_money(self):
        result = self.MoneyClass('2', 'XXX') + self.MoneyClass('2', 'XXX')
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_add_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') + self.MoneyClass(2, 'BBB')

    def test_add_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(0, 'XXX') + None

    def test_radd_int(self):
        result = 2 + self.MoneyClass('2', 'XXX')
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_sub_int(self):
        result = self.MoneyClass('2', 'XXX') - 2
        self.assertEqual(result, self.MoneyClass('0', 'XXX'))

    def test_sub_decimal(self):
        result = self.MoneyClass('2', 'XXX') - Decimal(2)
        self.assertEqual(result, self.MoneyClass('0', 'XXX'))

    def test_sub_money(self):
        result = self.MoneyClass('2', 'XXX') - self.MoneyClass('2', 'XXX')
        self.assertEqual(result, self.MoneyClass('0', 'XXX'))

    def test_sub_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') - self.MoneyClass(2, 'BBB')

    def test_sub_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(0, 'XXX') - None

    def test_rsub_int(self):
        result = 0 - self.MoneyClass('2', 'XXX')
        self.assertEqual(result, self.MoneyClass('-2', 'XXX'))

    def test_mul_int(self):
        result = self.MoneyClass('2', 'XXX') * 2
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_mul_decimal(self):
        result = self.MoneyClass('2', 'XXX') * Decimal(2)
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_mul_money(self):
        with self.assertRaises(TypeError):
            self.MoneyClass('2', 'XXX') * self.MoneyClass('2', 'XXX')

    def test_mul_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(0, 'XXX') * None

    def test_rmul_int(self):
        result = 2 * self.MoneyClass('2', 'XXX')
        self.assertEqual(result, self.MoneyClass('4', 'XXX'))

    def test_truediv_int(self):
        result = self.MoneyClass('2.99', 'XXX') / 2
        self.assertEqual(result, self.MoneyClass('1.495', 'XXX'))

    def test_truediv_decimal(self):
        result = self.MoneyClass('2.99', 'XXX') / Decimal(2)
        self.assertEqual(result, self.MoneyClass('1.495', 'XXX'))

    def test_truediv_money(self):
        result = self.MoneyClass('2', 'XXX') / self.MoneyClass('2', 'XXX')
        self.assertEqual(result, Decimal('1'))

    def test_truediv_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass(2, 'AAA') / self.MoneyClass(2, 'BBB')

    def test_truediv_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(2, 'XXX') / None

    def test_truediv_zero(self):
        with self.assertRaises(ZeroDivisionError):
            self.MoneyClass(2, 'XXX') / 0

    def test_floordiv_number(self):
        result = self.MoneyClass('2.99', 'XXX') // 2
        self.assertEqual(result, self.MoneyClass('1', 'XXX'))

    def test_floordiv_money(self):
        result = self.MoneyClass('2.99', 'XXX') // self.MoneyClass('2', 'XXX')
        self.assertEqual(result, Decimal('1'))

    def test_floordiv_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            self.MoneyClass('2.99', 'AAA') // self.MoneyClass('2', 'BBB')

    def test_floordiv_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(2, 'XXX') // None

    def test_floordiv_zero(self):
        with self.assertRaises(ZeroDivisionError):
            self.MoneyClass(2, 'XXX') // 0

    def test_mod_number(self):
        result = self.MoneyClass('2.99', 'XXX') % 2
        self.assertEqual(result, self.MoneyClass('0.99', 'XXX'))

    def test_mod_money(self):
        with self.assertRaises(TypeError):
            self.MoneyClass('2.99', 'XXX') % self.MoneyClass('2', 'XXX')

    def test_mod_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(2, 'XXX') % None

    def test_mod_zero(self):
        with self.assertRaises(ZeroDivisionError):
            self.MoneyClass(2, 'XXX') % 0

    def test_divmod_number(self):
        whole, remainder = divmod(self.MoneyClass('2.99', 'XXX'), 2)
        self.assertEqual(whole, self.MoneyClass('1', 'XXX'))
        self.assertEqual(remainder, self.MoneyClass('0.99', 'XXX'))

    def test_divmod_money(self):
        whole, remainder = divmod(
            self.MoneyClass(
                '2.99', 'XXX'), self.MoneyClass(
                '2', 'XXX'))
        self.assertEqual(whole, Decimal('1'))
        self.assertEqual(remainder, Decimal('0.99'))

    def test_divmod_money_different_currency(self):
        with self.assertRaises(CurrencyMismatch):
            divmod(self.MoneyClass('2.99', 'AAA'), self.MoneyClass('2', 'BBB'))

    def test_divmod_none(self):
        with self.assertRaises(TypeError):
            divmod(self.MoneyClass(2, 'XXX'), None)

    def test_divmod_zero(self):
        with self.assertRaises(ZeroDivisionError):
            divmod(self.MoneyClass(2, 'XXX'), 0)

    def test_pow_number(self):
        result = self.MoneyClass('3', 'XXX') ** 2
        self.assertEqual(result, self.MoneyClass('9', 'XXX'))

    def test_pow_money(self):
        with self.assertRaises(TypeError):
            self.MoneyClass('3', 'XXX') ** self.MoneyClass('2', 'XXX')

    def test_pow_none(self):
        with self.assertRaises(TypeError):
            self.MoneyClass(0, 'XXX') ** None

    def test_neg(self):
        result = -self.MoneyClass('2.99', 'XXX')
        self.assertEqual(result, self.MoneyClass('-2.99', 'XXX'))

    def test_pos(self):
        result = +self.MoneyClass('2.99', 'XXX')
        self.assertEqual(result, self.MoneyClass('2.99', 'XXX'))

    def test_abs(self):
        result = abs(self.MoneyClass('-2.99', 'XXX'))
        self.assertEqual(result, self.MoneyClass('2.99', 'XXX'))

    def test_int(self):
        self.assertEqual(int(self.MoneyClass('-2.99', 'XXX')), -2)
        self.assertEqual(int(self.MoneyClass('2.99', 'XXX')), 2)

    def test_float(self):
        self.assertEqual(float(self.MoneyClass('-2.99', 'XXX')), -2.99)
        self.assertEqual(float(self.MoneyClass('2.99', 'XXX')), 2.99)

    def test_round(self):
        self.assertEqual(
            round(
                self.MoneyClass(
                    '-1.49',
                    'XXX')),
            self.MoneyClass(
                '-1',
                'XXX'))
        self.assertEqual(
            round(
                self.MoneyClass(
                    '1.50', 'XXX')), self.MoneyClass(
                '2', 'XXX'))
        self.assertEqual(
            round(
                self.MoneyClass(
                    '1.234', 'XXX'), 2), self.MoneyClass(
                '1.23', 'XXX'))


class UnaryOperationsReturnNewMixin(object):
    def test_pos(self):
        self.assertIsNot(+self.money, self.money)

    def test_abs(self):
        self.assertIsNot(abs(self.money), self.money)

    def test_round(self):
        self.assertIsNot(round(self.money), self.money)


class LeftmostTypePrevailsMixin(object):
    def test_add(self):
        result = self.money + self.other_money
        self.assertEqual(result.__class__, self.MoneyClass)

    def test_add_other(self):
        result = self.other_money + self.money
        self.assertEqual(result.__class__, self.MoneySubclass)

    def test_sub(self):
        result = self.money - self.other_money
        self.assertEqual(result.__class__, self.MoneyClass)

    def test_sub_other(self):
        result = self.other_money - self.money
        self.assertEqual(result.__class__, self.MoneySubclass)



"""
Python money class with optional CLDR-backed locale-aware formatting
and an extensible currency exchange solution.

This is a modernized Python 3.9+ version of the original 'money' library
by Carlos Palol (https://github.com/carlospalol/money).
"""
from .money import Money, XMoney  # noqa: F401
from .exchange import xrates  # noqa: F401


# RADAR: version
__version__ = '2.0.0'



"""
Money exceptions
"""


class MoneyException(Exception):
    """Generic money exception"""

    pass


class CurrencyMismatch(MoneyException, ValueError):
    """Invalid operation between money objects of different currencies"""

    def __init__(self, a, b, operation):
        msg = (
            "unsupported operation between money in '{}' and '{}': '{}'. "
            "Use XMoney for automatic currency conversion."
        ).format(a, b, operation)
        super().__init__(msg)


class InvalidOperandType(MoneyException, TypeError):
    def __init__(self, operand, operation):
        msg = (
            "unsupported operation between Money and '{}': '{}'. This "
            "operation can only be performed with another Money object. "
            "You can access the amount of a money object 'foo'"
            " with 'foo.amount' (decimal.Decimal)."
        ).format(type(operand), operation)
        super().__init__(msg)


class ExchangeError(MoneyException):
    """Generic exception related to exchange rates"""

    pass


class ExchangeBackendNotInstalled(ExchangeError):
    """No backend installed yet"""

    def __init__(self):
        msg = "use e.g. money.xrates.install('money.exchange.SimpleBackend')"
        super().__init__(msg)


class ExchangeRateNotFound(ExchangeError):
    """A rate/quotation was not returned by the backend"""

    def __init__(self, backend, a, b):
        msg = "rate not found in backend '{}': {}/{}".format(backend, a, b)
        super().__init__(msg)



"""
Money doctests as unittest Suite
"""

import unittest

FILES = (
    # Skip old README.rst doctests - they contain outdated examples
    # '../../README.rst',
)


def load_tests(loader, tests, pattern):
    # Return empty test suite since README examples are outdated
    return unittest.TestSuite()



"""
Money classes
"""

import decimal
import re
from packaging.version import Version

from .exceptions import CurrencyMismatch, ExchangeRateNotFound, InvalidOperandType
from .exchange import xrates

__all__ = ["Money", "XMoney"]

BABEL_AVAILABLE = False
BABEL_VERSION = None
REGEX_CURRENCY_CODE = re.compile("^[A-Z]{3}$")
LC_NUMERIC = None

try:
    import babel
    import babel.numbers

    BABEL_AVAILABLE = True
    BABEL_VERSION = Version(babel.__version__)
    LC_NUMERIC = babel.default_locale("LC_NUMERIC")
except ImportError:
    pass


class Money(object):
    """Money class with a decimal amount and a currency"""

    def __init__(self, amount="0", currency=None):
        try:
            self._amount = decimal.Decimal(amount)
        except decimal.InvalidOperation as ex:
            raise ValueError(
                "amount value could not be "
                "converted to Decimal(): '{}'".format(amount)
            ) from ex
        if currency in [None, False, ""]:
            raise ValueError("invalid currency value: '{}'".format(currency))
        if not REGEX_CURRENCY_CODE.match(currency):
            raise ValueError(
                "currency not in ISO 4217 format: '{}'".format(currency)
            )
        self._currency = currency

    @property
    def amount(self):
        return self._amount

    @property
    def currency(self):
        return self._currency

    def __hash__(self):
        return hash((self._amount, self._currency))

    def __repr__(self):
        return "{} {}".format(self._currency, self._amount)

    def __str__(self):
        return "{} {:,.2f}".format(self._currency, self._amount)

    def __lt__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, "<")
        elif other.currency != self._currency:
            raise CurrencyMismatch(self._currency, other.currency, "<")
        else:
            return self._amount < other.amount

    def __le__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, "<=")
        elif other.currency != self._currency:
            raise CurrencyMismatch(self._currency, other.currency, "<=")
        else:
            return self._amount <= other.amount

    def __eq__(self, other):
        if isinstance(other, Money):
            return (self._amount == other.amount) and (self._currency == other.currency)
        return False

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, ">")
        elif other.currency != self._currency:
            raise CurrencyMismatch(self._currency, other.currency, ">")
        else:
            return self._amount > other.amount

    def __ge__(self, other):
        if not isinstance(other, Money):
            raise InvalidOperandType(other, ">=")
        elif other.currency != self._currency:
            raise CurrencyMismatch(self._currency, other.currency, ">=")
        else:
            return self._amount >= other.amount

    def __bool__(self):
        """
        Considering Money a numeric type (on ``amount``):

        bool(Money(2, 'XXX')) --> True
        bool(Money(0, 'XXX')) --> False
        """
        return bool(self._amount)

    def __add__(self, other):
        if isinstance(other, Money):
            if other.currency != self._currency:
                raise CurrencyMismatch(self._currency, other.currency, "+")
            other = other.amount
        amount = self._amount + other
        return self.__class__(amount, self._currency)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Money):
            if other.currency != self._currency:
                raise CurrencyMismatch(self._currency, other.currency, "-")
            other = other.amount
        amount = self._amount - other
        return self.__class__(amount, self._currency)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __mul__(self, other):
        if isinstance(other, Money):
            raise TypeError("multiplication is unsupported between two money objects")
        amount = self._amount * other
        return self.__class__(amount, self._currency)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Money):
            if other.currency != self._currency:
                raise CurrencyMismatch(self._currency, other.currency, "/")
            elif other.amount == 0:
                raise ZeroDivisionError()
            return self._amount / other.amount
        else:
            if other == 0:
                raise ZeroDivisionError()
            amount = self._amount / other
            return self.__class__(amount, self._currency)

    def __floordiv__(self, other):
        if isinstance(other, Money):
            if other.currency != self._currency:
                raise CurrencyMismatch(self._currency, other.currency, "//")
            elif other.amount == 0:
                raise ZeroDivisionError()
            return self._amount // other.amount
        else:
            if other == 0:
                raise ZeroDivisionError()
            amount = self._amount // other
            return self.__class__(amount, self._currency)

    def __mod__(self, other):
        if isinstance(other, Money):
            raise TypeError(
                "modulo is unsupported between two '{}' "
                "objects".format(self.__class__.__name__)
            )
        if other == 0:
            raise ZeroDivisionError()
        amount = self._amount % other
        return self.__class__(amount, self._currency)

    def __divmod__(self, other):
        if isinstance(other, Money):
            if other.currency != self._currency:
                raise CurrencyMismatch(self._currency, other.currency, "divmod")
            elif other.amount == 0:
                raise ZeroDivisionError()
            return divmod(self._amount, other.amount)
        else:
            if other == 0:
                raise ZeroDivisionError()
            whole, remainder = divmod(self._amount, other)
            return (
                self.__class__(whole, self._currency),
                self.__class__(remainder, self._currency),
            )

    def __pow__(self, other):
        if isinstance(other, Money):
            raise TypeError(
                "power operator is unsupported between two '{}' "
                "objects".format(self.__class__.__name__)
            )
        amount = self._amount**other
        return self.__class__(amount, self._currency)

    def __neg__(self):
        return self.__class__(-self._amount, self._currency)

    def __pos__(self):
        return self.__class__(+self._amount, self._currency)

    def __abs__(self):
        return self.__class__(abs(self._amount), self._currency)

    def __int__(self):
        return int(self._amount)

    def __float__(self):
        return float(self._amount)

    def __round__(self, ndigits=0):
        return self.__class__(round(self._amount, ndigits), self._currency)

    def __composite_values__(self):
        return self._amount, self._currency

    def to(self, currency):
        """Return equivalent money object in another currency"""
        if currency == self._currency:
            return self
        rate = xrates.quotation(self._currency, currency)
        if rate is None:
            raise ExchangeRateNotFound(xrates.backend_name, self._currency, currency)
        amount = self._amount * rate
        return self.__class__(amount, currency)

    def format(
        self,
        locale=LC_NUMERIC,
        pattern=None,
        currency_digits=True,
        format_type="standard",
    ):
        """
        Return a locale-aware, currency-formatted string.

        This method emulates babel.numbers.format_currency().

        A specific locale identifier (language[_territory]) can be passed,
        otherwise the system's default locale will be used. A custom
        formatting pattern of the form "¤#,##0.00;(¤#,##0.00)"
        (positive[;negative]) can also be passed, otherwise it will be
        determined from the locale and the CLDR (Unicode Common Locale Data
        Repository) included with Babel.

        >>> m = Money('1234.567', 'EUR')
        >>> m.format() # assuming the system's locale is 'en_US'
        €1,234.57
        >>> m.format('de_DE') # German formatting
        1.234,57 €
        >>> m.format('de', '#,##0 ¤') # German formatting (short), no cents
        1.235 €
        >>> m.format(pattern='#,##0.00 ¤¤¤') # Default locale, full name
        1,235.57 euro

        Learn more about this formatting syntaxis at:
        http://www.unicode.org/reports/tr35/tr35-numbers.html
        """
        if BABEL_AVAILABLE:
            return babel.numbers.format_currency(
                self._amount,
                self._currency,
                format=pattern,
                locale=locale,
                currency_digits=currency_digits,
                format_type=format_type,
            )
        else:
            raise NotImplementedError(
                "formatting requires Babel (https://pypi.python.org/pypi/Babel)"
            )

    @classmethod
    def loads(cls, s):
        """Parse from a string representation (repr)"""
        try:
            currency, amount = s.strip().split(" ")
            return cls(amount, currency)
        except ValueError as err:
            raise ValueError(
                "failed to parse string " " '{}': {}".format(s, err)
            ) from None


class XMoney(Money):
    """Money subclass with implicit currency conversion"""

    def __add__(self, other):
        if isinstance(other, Money):
            other = other.to(self._currency)
        return super().__add__(other)

    def __sub__(self, other):
        if isinstance(other, Money):
            other = other.to(self._currency)
        return super().__sub__(other)

    def __truediv__(self, other):
        if isinstance(other, Money):
            other = other.to(self._currency)
        return super().__truediv__(other)

    def __floordiv__(self, other):
        if isinstance(other, Money):
            other = other.to(self._currency)
        return super().__floordiv__(other)

    def __divmod__(self, other):
        if isinstance(other, Money):
            other = other.to(self._currency)
        return super().__divmod__(other)
